@@ -4,6 +4,7 @@
 import abc
 import functools
 import itertools
+import logging
 import pathlib
 import re
 import shelve
@@ -13,6 +14,12 @@ import sys
 import datetime
 import numpy
 
+from . import tools
+
+class InvalidFileError(Exception):
+    """Raised when the requested information cannot be obtained from the file
+    """
+
 class Dataset(metaclass=abc.ABCMeta):
     """Represents a dataset.
 
@@ -35,11 +42,21 @@ class Dataset(metaclass=abc.ABCMeta):
         
         Name for the dataset.  May be used for logging purposes and so.
 
+    - aliases::
+
+        Aliases for field.  Dictionary can be useful if you want to
+        programmatically loop through the same field for many different
+        datasets, but they are named differently.  Aliases known to be
+        used::
+
+        - ch4_profile
+
     """
 
     start_date = None
     end_date = None
     name = ""
+    aliases = {}
 
     def __init__(self, **kwargs):
         for (k, v) in kwargs.items():
@@ -75,23 +92,29 @@ class Dataset(metaclass=abc.ABCMeta):
         """
         raise NotImplementedError()
 
-    def read_period(self, start=datetime.datetime.min,
-                          end=datetime.datetime.max,
-                          onerror="skip"):
+    def read_period(self, start=None,
+                          end=None,
+                          onerror="skip",
+                          fields="all"):
         """Read all granules between start and end, in bulk.
 
         :param datetime start: Starting time, None for any time
         :param datetime end: Ending time, None for any time
         :param str onerror: What to do with unreadable files.  Defaults to
             "skip", can be set to "raise".
+        :param list fields: List of fields to read, or "all" (default).
         :returns: Masked array with all data in period.
         """
 
+        start = start or self.start_date
+        end = end or self.end_date
+
         contents = []
         for gran in self.find_granules(start, end):
             try:
-                cont = self.read(str(gran))
-            except ValueError as exc:
+                logging.info("Reading {!s}".format(gran))
+                cont = self.read(str(gran), fields=fields)
+            except (OSError, ValueError) as exc:
                 if onerror == "skip":
                     print("Could not read file {}: {}".format(
                         gran, exc.args[0], file=sys.stderr))
@@ -101,42 +124,50 @@ class Dataset(metaclass=abc.ABCMeta):
             else:
                 contents.append(cont)
         # retain type of first result, ordinary array of masked array
-        return (numpy.ma.concatenate 
+        arr = (numpy.ma.concatenate 
             if isinstance(contents[0], numpy.ma.MaskedArray)
             else numpy.concatenate)(contents)
+        return arr if fields == "all" else arr[fields]
 #        return numpy.ma.concatenate(list(
 #            self.read(f) for f in self.find_granules(start, end)))
 
-    def read_all(self):
+    def read_all(self, fields="all", onerror="skip"):
         """Read all data in one go.
 
         Warning: for some datasets, this may cause memory problems.
+
+        :param fields: List of fields to read, or "all" (default)
+        :param onerror: What to do in case of an error.  See
+            `read_period`.
         """
 
-        return self.read_period()
+        return self.read_period(fields=fields)
             
     @abc.abstractmethod
-    def _read(self, f):
+    def _read(self, f, fields="all"):
         """Read granule in file, low-level
 
         Shall return an ndarray with at least the fields lat, lon, time.
 
         :param str f: Path to file
+        :param fields: List of fields to read, or "all" (default)
         """
 
         raise NotImplementedError()
 
-    @functools.lru_cache(maxsize=10)
-    def read(self, f):
+    #@functools.lru_cache(maxsize=10)
+    @tools.mutable_cache(maxsize=10)
+    def read(self, f=None, fields="all"):
         """Read granule in file and do some other fixes
 
         Uses self._read.  Do not override, override _read instead.
 
         :param str f: Path to file
+        :param list fields: Fields to read or "all" (default)
         """
         if isinstance(f, pathlib.PurePath):
             f = str(f)
-        return self._read(f)
+        return self._read(f) if f is not None else self._read()
 
 
 class SingleFileDataset(Dataset):
@@ -156,6 +187,9 @@ class SingleFileDataset(Dataset):
                                    end=datetime.datetime.max):
         yield from self.find_granules(start, end)
 
+    def get_times_for_granule(self, gran=None):
+        return (self.start_date, self.end_date)
+
 class MultiFileDataset(Dataset):
     """Represents a dataset where measurements are spread over multiple
     files.
@@ -184,7 +218,7 @@ class MultiFileDataset(Dataset):
             number, etc.  For time identification, relevant fields are
             contained in MultiFileDataset.date_info, where each field also
             exists in a version with "_end" appended.
-            MultiFileDataset.re_info contains all recognised fields.
+            MultiFileDataset.refields contains all recognised fields.
 
             If any *_end fields are found, the ending time is equal to the
             beginning time with any *_end fields replaced.  If no *_end
@@ -224,8 +258,13 @@ class MultiFileDataset(Dataset):
                 not isinstance(getattr(self, attr), pathlib.PurePath)):
                 setattr(self, attr, pathlib.Path(getattr(self, attr)))
         if self.granule_cache_file is not None:
-            self._granule_start_times = shelve.open(str(self.basedir /
-                self.granule_cache_file), protocol=4)
+            p = str(self.basedir / self.granule_cache_file)
+            try:
+                self._granule_start_times = shelve.open(p, protocol=4)
+            except OSError:
+                logging.error(("Unable to open granule file {} RW.  "
+                               "Opening read-only.").format(p))
+                self._granule_start_times = shelve.open(p, flag='r')
         else:
             self._granule_start_times = {}
         if self.re is not None:
@@ -317,13 +356,22 @@ class MultiFileDataset(Dataset):
         d_end = (dt_end.date()
                 if isinstance(dt_end, datetime.datetime) 
                 else dt_end)
+        logging.info(("Searching for granules between {!s} and {!s}"
+                      "for {}").format(dt_start, dt_end, self.name))
         for (timeinfo, subdir) in self.iterate_subdirs(d_start, d_end):
             if subdir.exists() and subdir.is_dir():
+                logging.debug("Searching directory {!s}".format(subdir))
                 for child in subdir.iterdir():
                     m = self._re.fullmatch(child.name)
                     if m is not None:
-                        (g_start, g_end) = self.get_times_for_granule(child,
-                            **timeinfo)
+                        try:
+                            (g_start, g_end) = self.get_times_for_granule(child,
+                                **timeinfo)
+                        except InvalidFileError as e:
+                            logging.error(
+                                "Skipping {!s}.  Problem: {}".format(
+                                    child, e.args[0]))
+                            continue
                         if g_end > dt_start and g_start < dt_end:
                             yield child
             
@@ -374,6 +422,8 @@ class MultiFileDataset(Dataset):
         May take hints for year, month, day, hour, minute, second, and
         their endings, according to self.date_fields
         """
+        if not isinstance(p, pathlib.PurePath):
+            p = pathlib.PurePath(p)
         if str(p) in self._granule_start_times.keys():
             (start, end) = self._granule_start_times[str(p)]
         else:
@@ -410,6 +460,8 @@ class MultiFileDataset(Dataset):
     # and only if starting/ending times cannot be determined from the filename
     def get_time_from_granule_contents(self, p):
         """Get datetime objects for beginning and end of granule
+
+        If it returns None, then use same as start time.
         """
         raise ValueError(
             ("To determine starting and end-times for a {0} dataset, "
@@ -423,8 +475,10 @@ class SingleMeasurementPerFileDataset(MultiFileDataset):
     An example of this would be ACE, or some RO datasets.
     """
 
+    granule_duration = datetime.timedelta(0)
+
     @abc.abstractmethod
-    def read_single(self, p):
+    def read_single(self, p, fields="all"):
         """Read a single measurement from a single file.
 
         Shall take one argument (a path object) and return a tuple with
@@ -432,14 +486,19 @@ class SingleMeasurementPerFileDataset(MultiFileDataset):
         latitude, longitude, time.
         """
 
-    def _read(self, p):
+    _head_dtype = dict(CH4_total="<f4")
+    def _read(self, p, fields="all"):
         """Reads a single measurement converted to ndarray
         """
 
-        (head, body) = self.read_single(p)
+        (head, body) = self.read_single(p, fields=fields)
 
-        dt = [s+body.shape for s in body.dtype.descr]
+        dt = [(s+body.shape if len(s)==2 else (s[0], s[1], s[2]+body.shape))
+                for s in body.dtype.descr]
         dt.extend([("lat", "f8"), ("lon", "f8"), ("time", "M8[s]")])
+        dt.extend([(s, self._head_dtype[s])
+                for s in (head.keys() & self._head_dtype.keys())
+                if s not in {"lat", "lon", "time"}])
         # This fails. https://github.com/numpy/numpy/issues/4583
         #D = numpy.ma.empty(1, dt)
         D = numpy.empty(1, dt)
@@ -450,6 +509,10 @@ class SingleMeasurementPerFileDataset(MultiFileDataset):
         for nm in {"lat", "lon", "time"}:
             D[nm] = head[nm]
 
+        for nm in head.keys() & D.dtype.names:
+            if nm not in {"lat", "lon", "time"}:
+                D[nm] = head[nm]
+
         return D
 
 class HomemadeDataset(MultiFileDataset):
@@ -460,8 +523,28 @@ class HomemadeDataset(MultiFileDataset):
     # dummy implementation for abstract methods, so that I can test other
     # things
 
-    def _read(self, f):
+    def _read(self, f, fields="all"):
         raise NotImplementedError()
 
     def find_granules(self, start, end):
         raise StopIteration()
+
+#    @abc.abstractmethod
+#    def quicksave(self, f):
+#        """Quick save to file
+#
+#        :param str f: File to save to
+#        """
+
+
+class ProfileDataset(Dataset):
+    """Abstract superclass with stuff for anything having profiles
+    """
+
+    @abc.abstractmethod
+    def get_z(self, dt):
+        """Get z-profile in metre.
+
+        Takes as argument a single measurement, with the dtype returned by
+        its own reading routine.
+        """
