@@ -17,7 +17,7 @@ from .constants import ppm as PPM
 
 HECTO = 100
 
-class TansoFTS(dataset.SingleFileDataset):
+class TansoFTS(dataset.SingleFileDataset, dataset.ProfileDataset):
 
 #    start_date = datetime.datetime(2010, 3, 23, 2, 24, 54, 210)
 #    end_date = datetime.datetime(2010, 10, 31, 20, 34, 50, 814)
@@ -29,6 +29,7 @@ class TansoFTS(dataset.SingleFileDataset):
         700, 600, 500, 400, 300, 250, 200, 150, 100, 70, 50, 30, 20,
         10])*HECTO # hPa -> Pa
     p_for_interp_profile = None
+    aliases = {"CH4_profile": "ch4_profile_raw"}
 
     @classmethod
     def vmr_to_column_density(cls, data):
@@ -74,12 +75,12 @@ class TansoFTS(dataset.SingleFileDataset):
 
             nd = physics.vmr2nd(
                 data["ch4_profile_raw"][n, :].data[
-                    ~data["ch4_profile_raw"][n, :].mask]*PPM,
+                    ~data["ch4_profile_raw"][n, :].mask],#*PPM,
                 T_interp, p)
 
             nd_e = physics.vmr2nd(
                 data["ch4_profile_raw_e"][n, :].data[
-                    ~data["ch4_profile_raw_e"][n, :].mask]*PPM,
+                    ~data["ch4_profile_raw_e"][n, :].mask],#*PPM,
                 T_interp, p)
 
             #z = physics.p2z_oversimplified(p)
@@ -111,11 +112,10 @@ class TansoFTS(dataset.SingleFileDataset):
         return (ncd, ncd_e)
 
 
-
     # implementation of abstract methods
 
-    def _read(self, path=None):
-        """Read Tanso FTP granule.  Currently hardcoded for CH4 raw profiles.
+    def _read(self, path=None, fields="all"):
+        """Read Tanso FTS granule.  Currently hardcoded for CH4 raw&interp.
         
         """
         if path is None:
@@ -139,13 +139,17 @@ class TansoFTS(dataset.SingleFileDataset):
             else:
                 self.p_for_interp_profile = p
             #D["p"] = h5f["Data"]["originalProfile"]["pressure"]
-            D["ch4_profile_interp"] = h5f["Data"]["interpolatedProfile"]["CH4Profile"]
-            D["ch4_profile_raw"] = h5f["Data"]["originalProfile"]["CH4Profile"]
-            D["ch4_profile_raw_e"] = h5f["Data"]["originalProfile"]["CH4ProfileError"]
+            D["ch4_profile_interp"] = (h5f["Data"]["interpolatedProfile"]
+                ["CH4Profile"][:]*PPM)
+            D["ch4_profile_raw"] = (h5f["Data"]["originalProfile"]
+                ["CH4Profile"][:]*PPM)
+            D["ch4_profile_raw_e"] = (h5f["Data"]["originalProfile"]
+                ["CH4ProfileError"][:]*PPM)
             D["p_raw"] = h5f["Data"]["originalProfile"]["pressure"]
             D["T"] = h5f["scanAttribute"]["referenceData"]["temperatureProfile"]
             D["h2o"] = h5f["scanAttribute"]["referenceData"]["waterVaporProfile"]
             D["p0"] = h5f["scanAttribute"]["referenceData"]["surfacePressure"]
+            D["id"] = h5f["scanAttribute"]["scanID"]
 
 
             A = numpy.empty(shape=time_raw.size,
@@ -158,7 +162,77 @@ class TansoFTS(dataset.SingleFileDataset):
                 A.mask[k][A.data[k]<0] = True
         A["p0"] *= HECTO
 
-        return A
+        return A if fields=="all" else A[fields]
+
+    def get_z(self, meas): 
+#            if data["T"].mask.size == 1:
+#                T = data["T"][n, ::-1].data
+#            else:
+#                T = data["T"][n, ::-1].data[~data["T"][n, ::-1].mask]
+#
+#            if data["h2o"].mask.size == 1:
+#                h2o = data["h2o"][n, ::-1].data
+#            else:
+#                h2o = data["h2o"][n, ::-1].data[~data["h2o"][n, ::-1].mask]
+
+
+        # Can't use masked arrays due to bug:
+        # https://github.com/numpy/numpy/issues/2972
+        # Use workaround instead.
+        if isinstance(meas, numpy.ma.core.MaskedArray):
+            meas = meas.data
+
+        valid = meas["p_raw"] > 0
+
+        T_interpolator = scipy.interpolate.interp1d(
+            self.p_for_T_profile[::-1],
+            meas["T"][::-1])# should be increasing
+
+        h2o_interpolator = scipy.interpolate.interp1d(
+            self.p_for_T_profile[::-1],
+            meas["h2o"][::-1])
+        
+#            p = data["p_raw"][n, :].data[~data["p_raw"][n, :].mask]*HECTO
+        p = meas["p_raw"] * HECTO
+        p = p[valid]
+
+        T_interp = T_interpolator(p)
+        h2o_interp = h2o_interpolator(p)
+
+#            nd = physics.vmr2nd(
+#                data["ch4_profile_raw"][n, :].data[
+#                    ~data["ch4_profile_raw"][n, :].mask]*PPM,
+#                T_interp, p)
+#
+#            nd_e = physics.vmr2nd(
+#                data["ch4_profile_raw_e"][n, :].data[
+#                    ~data["ch4_profile_raw_e"][n, :].mask]*PPM,
+#                T_interp, p)
+
+        #z = physics.p2z_oversimplified(p)
+        if meas["p0"] > p[0]:
+            # need to add dummy p[0] < p to satisfy p2z
+            #
+            # this is inaccurate, but has little effect on the
+            # *relative* location of other levels (<1 m), so it is an
+            # acceptable source of error
+            dummy = True
+            pp = numpy.hstack((meas["p0"], p))
+            Tp = numpy.hstack((287, T_interp))
+            # although putting h2o-surface to a low value is an EXTREMELY
+            # bad approximation, it /still/ doesn't matter for the
+            # relative thicknesses higher up
+            h2op = numpy.hstack((h2o_interp[0], h2o_interp))
+        else:
+            dummy = False
+            (pp, Tp, h2op) = (p, T_interp, h2o_interp)
+        #z = physics.p2z_hydrostatic(p, T_interp, h2o_interp,
+        z = physics.p2z_hydrostatic(pp, Tp, h2op*PPM,
+            p0=meas["p0"],
+            z0=meas["z0"],
+            lat=meas["lat"])[dummy:]
+        return z
+
 
 class NDACCAmes(dataset.MultiFileDataset):
     """NDACC Ames-style file
@@ -195,7 +269,7 @@ class NDACCAmes(dataset.MultiFileDataset):
 #        return (datetime.datetime(y1, m1, d1, 0, 0, 0),
 #                datetime.datetime(y2, m2, d2, 23, 59, 59))
              
-    def _read(self, path):
+    def _read(self, path, fields="all"):
         """Read Bruker data in NDACC Gaines format
 
         Returns a masked array
@@ -247,7 +321,7 @@ class NDACCAmes(dataset.MultiFileDataset):
                     new["time"] = dt
                     for field in v.dtype.names:
                         new[field] = v[field]
-                    L.append(new)
+                    L.append(new if fields=="all" else new[fields])
 
         # now I have an array with fractional years as the time-axis.
         # I want to have datetime64
@@ -283,8 +357,9 @@ class ACEFTS(dataset.SingleMeasurementPerFileDataset):
     """
 #    basedir = "/home/gerrit/sshfs/glacier/data/1/gholl/data/ACE"
     subdir = "{year:04d}-{month:02d}"
-    re = r"(?P<type>s[sr])(?P<orbit>\d{5})v3\.0\.asc"
+    re = r"(?P<type>s[sr])(?P<orbit>\d{5})v(?P<version>\d\.\d)\.asc"
     _time_format = "%Y-%m-%d %H:%M:%S"
+    aliases = {"CH4_profile": "CH4"}
 
     @staticmethod
     def read_header(fp):
@@ -301,9 +376,12 @@ class ACEFTS(dataset.SingleMeasurementPerFileDataset):
         for line in itertools.takewhile(isempty, fp):
             k, v = line.split("|")
             head[k.strip()] = v.strip()
+        if head == {}:
+            raise dataset.InvalidFileError(
+                "Unable to extract header from {0.name}.  Empty?".format(fp))
         return head
 
-    def read_single(self, f):
+    def read_single(self, f, fields="all"):
         with open(f) as fp:
             head = self.read_header(fp)
 
@@ -326,6 +404,9 @@ class ACEFTS(dataset.SingleMeasurementPerFileDataset):
                 for (i, name) in enumerate(names):
                     D[name][n] = vals[i]
 
+        # km -> m
+        D["z"] *= 1e3
+
         head["lat"] = float(head["latitude"])
         head["lon"] = float(head["longitude"])
         # make sure lons are in (-180, 180)
@@ -338,7 +419,7 @@ class ACEFTS(dataset.SingleMeasurementPerFileDataset):
         # decimal part (truncating it to the nearest second)
         head["time"] = datetime.datetime.strptime(
             head["date"].split(".")[0].split("+")[0], self._time_format)
-        return (head, D)
+        return (head, D if fields=="all" else D[fields])
         
     def get_time_from_granule_contents(self, p):
         """Get time from granule contents.
@@ -353,6 +434,13 @@ class ACEFTS(dataset.SingleMeasurementPerFileDataset):
         return tuple(datetime.datetime.strptime(
             head[m + "_time"][:-3], self._time_format)
             for m in ("start", "end"))
+
+    def get_z(self, meas):
+        m = meas["z"]
+        if m[-1] < 150: # oops, still in km
+            return m * 1e3
+        else:
+            return m
             
 
 def collect_values(fp, N, dtp):
